"""Class definition for DebPackageData."""

from __future__ import annotations

from hashlib import sha256

from choppr.apt_api import BinaryPackage


__all__ = ["DebPackageData"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class DebPackageData:
    """Class containing frequently accessed information and the originating package.

    Members:
        - name
        - package
        - version
    """

    def __init__(self, package: BinaryPackage) -> None:
        """Create an instance of DebPackageData.

        Arguments:
            package: The originating binary package
        """
        self.package = package
        self.name: str = package.package
        self.version: str = package.version
        self.provides: list[str] = package.provides

    def __eq__(self, other: object) -> bool:
        match other:
            case DebPackageData():
                return self.name == other.name and self.version == other.version
            case BinaryPackage():
                return self.name == str(other.package) and self.version == str(other.version)
            case _:
                return False

    def __hash__(self) -> int:
        sha = sha256()
        sha.update(self.name.encode())
        sha.update(self.version.encode())
        return int(sha.hexdigest(), 16)

    def __str__(self) -> str:
        return f"{self.name}-{self.version}"



"""ReleaseFile implementation."""

from __future__ import annotations

from choppr.apt_api._utils import get_value


__all__ = ["ReleaseFile"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class ReleaseFile:
    """Class that represents a Release file.

    # Arguments
    content (str): the content of the Release file
    """

    def __init__(self, content: str) -> None:
        """Initialize an instance of ReleaseFile.

        Arguments:
            content: _description_
        """
        self.__content = content.strip()

    @property
    def origin(self) -> str:
        """Get the value of Origin.

        Returns:
            str: Origin value
        """
        return str(get_value(self.__content, "Origin"))

    @property
    def label(self) -> str:
        """Get the value of Label.

        Returns:
            str: Label value
        """
        return str(get_value(self.__content, "Label"))

    @property
    def suite(self) -> str:
        """Get the value of Suite.

        Returns:
            str: Suite value
        """
        return str(get_value(self.__content, "Suite"))

    @property
    def version(self) -> str:
        """Get the value of Version.

        Returns:
            str: Version value
        """
        return str(get_value(self.__content, "Version"))

    @property
    def codename(self) -> str:
        """Get the value of Codename.

        Returns:
            str: Codename value
        """
        return str(get_value(self.__content, "Codename"))

    @property
    def date(self) -> str:
        """Get the value of Date.

        Returns:
            str: Date value
        """
        return str(get_value(self.__content, "Date"))

    @property
    def architectures(self) -> list[str]:
        """Get the value of Architectures.

        Returns:
            str: Architectures value
        """
        return str(get_value(self.__content, "Architectures")).split()

    @property
    def components(self) -> list[str]:
        """Get the value of Components.

        Returns:
            str: Components value
        """
        return str(get_value(self.__content, "Components")).split()

    @property
    def description(self) -> str:
        """Get the value of Description.

        Returns:
            str: Description value
        """
        return str(get_value(self.__content, "Description"))



"""Choppr refines the components in a Software Bill of Materials (SBOM).

It does not replace SBOM generation tools.  Mainly, Choppr analyses a build or runtime to verify
which components are used, and remove the SBOM components not used.  Starting with file accesses, it
works backwards from how an SBOM generation tool typically would.  For example SBOM generators use
the yum database to determine which packages yum installed.  Choppr looks at all the files accessed
and queries sources like yum to determine the originating package.
"""

from __future__ import annotations

import shutil

from typing import TYPE_CHECKING, Any

from hoppr import (
    BomAccess,
    ComponentType,
    HopprLoadDataError,
)
from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_process
from hoppr.models.types import PurlType
from hoppr.result import Result
from hoppr_cyclonedx_models.cyclonedx_1_6 import Scope

from choppr import __version__
from choppr.deb_handler import DebHandler
from choppr.rpm_handler import RpmHandler
from choppr.types import ChopprConfig, ChopprConfigModel, ChopprShares
from choppr.utils import get_component_type, log_header, output_list


if TYPE_CHECKING:
    from hoppr import Component, HopprContext, Sbom


__all__ = ["ChopprPlugin"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class ChopprPlugin(HopprPlugin):
    """Plugin implementation of Choppr to integrate with Hoppr."""

    supported_purl_types = ["rpm", "deb"]  # noqa: RUF012
    products = []  # noqa: RUF012
    bom_access = BomAccess.FULL_ACCESS

    def __init__(self, context: HopprContext, config: dict[Any, Any] | None = None) -> None:
        """Initialize plugin with Hoppr framework arguments (context and config).

        Arguments:
            context: Hoppr context
            config: Hoppr configuration
        """
        super().__init__(context, config)

        self.log = self.get_logger()
        self.log.flush_immed = True

        # Parse configuration
        self.valid_config = False
        try:
            plugin_config = ChopprConfig(ChopprConfigModel.parse_obj(config))
            ChopprShares(plugin_config, self.context, self.log)
            self.valid_config = True
        except (HopprLoadDataError, FileNotFoundError) as e:
            self.log.error(f"Invalid Configuration: {e}")  # noqa: TRY400

        self.file_provided_by: dict[str, str] = {}
        self.search_repositories = dict.fromkeys(PurlType, False)

        self.rpm = RpmHandler()
        self.deb = DebHandler()

    def get_version(self) -> str:  # noqa: PLR6301
        """Return the version of the Choppr plugin.

        Returns:
            str: Plugin version
        """
        return __version__

    ################################################################################################
    # Pre-Stage Process
    ################################################################################################

    @hoppr_process
    def pre_stage_process(self) -> Result:
        """Collect repository information.

        Returns:
            Result: Skip if the config is invalid, or the modified SBOM
        """
        if not self.valid_config:
            return Result.skip()

        self._output_excluded_components([])

        self._populate_repositories()
        self._populate_potential_packages()
        self._populate_required_packages()
        self._populate_nested_dependencies()

        choppd_sbom = self._filter_sbom()

        self._output_excluded_components(choppd_sbom.components)

        return Result.success(return_obj=choppd_sbom)

    def _populate_repositories(self) -> None:
        log_header("Populate Repositories")
        self.search_repositories[PurlType.RPM] = self.rpm.populate_repositories()
        self.search_repositories[PurlType.DEB] = self.deb.populate_repositories()

    def _populate_potential_packages(self, required_files: set[str] | None = None) -> None:
        log_header("Populate Potential Packages")
        using_strace_files = False

        if not required_files:
            using_strace_files = True
            required_files = ChopprShares.config.strace_files

        for file in required_files:
            if self.search_repositories[PurlType.RPM]:
                self.rpm.populate_potential_packages(file, using_strace_files)
            if self.search_repositories[PurlType.DEB]:
                self.deb.populate_potential_packages(file)

        if self.search_repositories[PurlType.RPM]:
            ChopprShares.log.info(f"Found {len(self.rpm.potential_packages)} potential rpm packages")

            with ChopprShares.config.cache_dir.joinpath("rpm", "potential_packages.txt").open("w") as output:
                output.writelines([f"{pkg}\n" for pkg in sorted(self.rpm.potential_packages, key=lambda p: p.name)])

        if self.search_repositories[PurlType.DEB]:
            ChopprShares.log.info(f"Found {len(self.deb.potential_packages)} potential deb packages")

            with ChopprShares.config.cache_dir.joinpath("deb", "potential_packages.txt").open("w") as output:
                output.writelines([f"{pkg}\n" for pkg in sorted(self.deb.potential_packages, key=lambda p: p.name)])

    def _populate_required_packages(self) -> None:
        log_header("Populate Required Packages")

        if self.search_repositories[PurlType.RPM]:
            self.rpm.populate_required_packages()
        if self.search_repositories[PurlType.DEB]:
            self.deb.populate_required_packages()

    def _populate_nested_dependencies(self) -> None:
        log_header("Populate Nested Dependencies")

        if self.search_repositories[PurlType.RPM]:
            self.rpm.populate_all_nested_dependencies()
        if self.search_repositories[PurlType.DEB]:
            self.deb.populate_all_nested_dependencies()

    def _filter_sbom(self) -> Sbom:
        log_header("Filter SBOM")
        choppd_sbom = ChopprShares.context.delivered_sbom.copy(deep=True)
        choppd_sbom.components.clear()

        for component in ChopprShares.context.delivered_sbom.components or []:
            component_id = f"{component.name}-{component.version}"

            if component_type := get_component_type(component):
                if (
                    component_type in ChopprShares.config.allowlist
                    and component in ChopprShares.config.allowlist[component_type]
                ):
                    component.scope = Scope.REQUIRED
                    ChopprShares.log.info(f"Component accepted by allowlist: {component_id}")
                    choppd_sbom.components.append(component)
                    continue
                if (
                    component_type in ChopprShares.config.denylist
                    and component in ChopprShares.config.denylist[component_type]
                ):
                    component.scope = Scope.EXCLUDED
                    ChopprShares.log.info(f"Component blocked by denylist: {component_id}")
                    continue

            if component_scope := self._get_component_scope(component):
                component.scope = component_scope
                match component_scope:
                    case Scope.REQUIRED:
                        ChopprShares.log.info(f"Component required: {component_id}")
                        choppd_sbom.components.append(component)
                    case Scope.EXCLUDED:
                        ChopprShares.log.info(f"Component not required: {component_id}")
            else:
                ChopprShares.log.warning(f"Unable to determine if component is required: {component_id}")

        if ChopprShares.config.delete_excluded:
            ChopprShares.log.debug("Deleted excluded components")
            return choppd_sbom
        return ChopprShares.context.delivered_sbom

    @staticmethod
    def _output_excluded_components(filtered_components: list[Component]) -> None:
        excluded_components_all = [
            c for c in ChopprShares.context.delivered_sbom.components if c not in filtered_components
        ]

        for purl_type in PurlType:
            output_format = ChopprShares.config.output_files.excluded_components[purl_type].component_format
            if excluded_components := [
                output_format.format(**c.dict()) for c in excluded_components_all if get_component_type(c) is purl_type
            ]:
                output_file = ChopprShares.config.output_files.excluded_components[purl_type].file
                output_file.parent.mkdir(parents=True, exist_ok=True)
                output_list(output_file, excluded_components)

    def _get_component_scope(self, component: Component) -> Scope | None:
        if ChopprShares.config.keep_essential_os_components and component.type is ComponentType.OPERATING_SYSTEM:
            return Scope.REQUIRED
        if not ChopprShares.config.allow_version_mismatch and not component.version:
            return None
        return next(
            (
                Scope.REQUIRED
                for package in self.rpm.required_packages
                | self.rpm.nested_dependencies
                | self.deb.required_packages
                | self.deb.nested_dependencies
                if (ChopprShares.config.allow_version_mismatch and str(package).startswith(component.name))
                or str(package) == f"{component.name}-{component.version}"
            ),
            Scope.EXCLUDED,
        )

    ################################################################################################
    # Post-Stage Process
    ################################################################################################

    @hoppr_process
    def post_stage_process(self) -> Result:
        """Perform cleanup tasks after running Choppr.

        Returns:
            Result: The updated SBOM with the unused packages removed
        """
        if not self.valid_config:
            return Result.skip()

        if ChopprShares.config.clear_cache:
            shutil.rmtree(ChopprShares.config.cache_dir)

        return Result.success()



"""PackagesFile implementation."""

from __future__ import annotations

import json

from typing import TYPE_CHECKING

from choppr.apt_api._utils import get_value
from choppr.apt_api.binary_package import BinaryPackage


if TYPE_CHECKING:
    from pathlib import Path


__all__ = ["PackagesFile"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class PackagesFile:
    """Class that represents a Packages file."""

    def __init__(self, packages_file: Path, content_file: Path, repo_url: str) -> None:
        """Initialize an instance of PackagesFile.

        Arguments:
            packages_file: The list of packages with their metadata
            content_file: The list of files internal to this specific package
            repo_url: The URL for the repository
        """
        self.packages: list[BinaryPackage] = []

        if packages_file.is_file():
            with packages_file.open(encoding="utf-16") as f:
                content = f.read().strip()

            with content_file.open() as f:
                file_list: dict[str, list[str]] = json.load(f)

            if packages_content := content.split("\n\n"):
                for package_content in packages_content:
                    if package_content:
                        name = str(get_value(package_content, "Package"))
                        if name in file_list:
                            bp = BinaryPackage(package_content, file_list[name], repo_url)
                        else:
                            bp = BinaryPackage(package_content, package_url=repo_url)

                        self.packages.append(bp)



"""Class definition for ChopprShares, a singleton class to share commonly access objects."""

from __future__ import annotations

from typing import TYPE_CHECKING

from hoppr import PurlType


if TYPE_CHECKING:
    from hoppr import Component, HopprContext, HopprLogger
    from typing_extensions import Self

    from choppr.types import ChopprConfig


__all__ = ["ChopprShares"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class ChopprShares:
    """Singleton class to hold commonly accessed Choppr objects.

    Members:
        - components
        - config
        - context
        - log

    Methods:
        - initialized
    """

    _instance: ChopprShares | None = None

    def __new__(cls, config: ChopprConfig, context: HopprContext, log: HopprLogger, *args, **kwargs) -> Self:
        """Initialize ChopprShares singleton.

        Arguments:
            config: Instance of ChopprConfig
            context: Instance of HopprContext
            log: Instance of HopprLogger

        Returns:
            Self: Singleton instance of ChopprShares
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)

            cls.components: dict[PurlType, list[Component]] = {
                purl_type: [
                    component
                    for component in context.delivered_sbom.components
                    if component.purl and component.purl.lower().startswith(f"pkg:{purl_type.value.lower()}")
                ]
                for purl_type in PurlType
            }
            cls.config: ChopprConfig = config
            cls.context: HopprContext = context
            cls.log: HopprLogger = log

            for purl_type, component_list in cls.components.items():
                cls.log.debug(f"{purl_type.name} component count: {len(component_list)}")
        return cls._instance

    @classmethod
    def initialized(cls) -> bool:
        """Check if the class has been initialized.

        Returns:
            bool: True if the class is initialized, otherwise False
        """
        return cls._instance is not None



from __future__ import annotations

import re


__all__ = ["get_list", "get_value"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


def get_value(content: str, key: str, default: str | None = None) -> str | None:
    """Extract a value from a Packages or Release file.

    Arguments:
        content: The content of the Packages/Release file
        key: The key to return the value for
        default: The default value to return if the key is not found

    Returns:
        str | None: Value for a given key
    """
    pattern = key + ": (.*)\n"
    return match[1] if (match := re.search(pattern, content)) else default


def get_list(content: str, key: str) -> list[str]:
    """Extract a list of values from a Packages or Release file.

    Arguments:
        content: The content of the Packages/Release file
        key: The key to return the value for

    Returns:
        list[str]: List of values for a given key
    """
    if field := get_value(content, key):
        return [d.strip() for d in field.split(",")]

    return []



"""Module containing functions to call strace and analyze output."""

from __future__ import annotations

import re

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from pathlib import Path


__all__ = ["get_files"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "Lockheed Martin Proprietary Information"


def get_files(strace_location: Path) -> set[str]:
    """Get the list of touched files from strace.

    Arguments:
        strace_location: Output file from strace (default strace_output.txt)

    Returns:
        set[str]: Set of touched files
    """
    with strace_location.open() as file:
        text = file.read()

    flags = re.MULTILINE | re.IGNORECASE

    dir_regex = (
        r"^.*?"
        r"(?:(?:mkdir).*?"
        r"(?:(?:\'|\")(?P<mkdir>.*?)(?:\'|\")))"
        r"|"
        r"(?:(?:(?:\'|\")(?P<dirname>.*?)(?:\'|\"))"
        r"(?=.*?(?:S_IFDIR|O_DIRECTORY))(?!.*(?:ENOENT|unfinished)))"
    )
    dir_matches = re.finditer(dir_regex, text, flags)
    directories: set[str] = {dm.group(g) for dm in dir_matches for g in ["dirname", "mkdir"]}

    regex = r"^.*?(?:(?:\'|\")(?P<filename>.*?)(?:\'|\"))(?!.*(?:S_IFDIR|ENOENT|O_DIRECTORY|unfinished))"
    return {match.group("filename") for match in re.finditer(regex, text, flags)} - directories



"""Constants used in Choppr."""

from __future__ import annotations

from typing import Final

from hoppr import PurlType


__all__ = [
    "DEB_HEADER_LENGTH",
    "DEFAULT_ARCH_DEB",
    "DEFAULT_ARCH_RPM",
    "DEFAULT_RECURSION_LIMIT",
    "LOG_HEADER_LENGTH",
]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


DEB_HEADER_LENGTH: Final[int] = 60
"""Length of headers in DEB package files."""

DEFAULT_ARCH_DEB: Final[str] = "amd64"
"""Default architecture for Debian repositories."""

DEFAULT_ARCH_RPM: Final[str] = "x86_64"
"""Default architecture for RPM repositories."""

DEFAULT_RECURSION_LIMIT: Final[int] = 10
"""Default depth to limit recursive functions to when used alongside the `limit_recursion` decorator."""

LOG_HEADER_LENGTH: Final[int] = 100
"""Length of headers in logs."""

COMPONENT_LIST_FORMATS: Final[dict[PurlType, str]] = {
    PurlType.DEB: "{name}={version}",
    PurlType.NPM: "{name}@{version}",
    PurlType.RPM: "{name}-{version}",
}



"""Module for handling DEB packages."""

from __future__ import annotations

from typing import TYPE_CHECKING, Final

from debian.debian_support import Version
from hoppr.models.manifest import Repository
from hoppr.models.types import PurlType
from pydantic import HttpUrl, parse_obj_as

from choppr import apt_api
from choppr.decorators import limit_recursion
from choppr.types import ChopprShares, DebPackageData
from choppr.utils import get_auth_and_verify, log_repo_pulls


if TYPE_CHECKING:
    from pathlib import Path

    from hoppr import Component

    from choppr.apt_api import BinaryPackage


__all__ = ["DebHandler"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


def _is_package_essential(package: BinaryPackage, build_essential: bool = False) -> bool:
    return (
        package.essential
        or (build_essential and package.build_essential)
        or (package.priority and package.priority.lower() == "required")
    )


class DebHandler:
    """Class to handle all DEB repository processing."""

    def __init__(self) -> None:
        self.allow_version_mismatch = ChopprShares.config.allow_version_mismatch
        self.cache_dir = ChopprShares.config.cache_dir / "deb"
        self.cache_dir.mkdir(parents=True, exist_ok=True)

        self.file_provided_by: dict[str, str] = {}
        self.repositories: apt_api.Sources | None = None
        self.potential_packages: set[DebPackageData] = set()
        self.required_packages: set[DebPackageData] = set()
        self.nested_dependencies: set[DebPackageData] = set()

    ####################################################################################################
    # Exported Methods
    ####################################################################################################

    def populate_repositories(self) -> bool:
        """Pull all of the metadata for DEB repositories provided in the config.

        Returns:
            bool: True when any repository was successfully pulled
        """
        if not (
            deb_repository_urls := [
                str(repo.url)
                for repo in parse_obj_as(list[Repository], ChopprShares.context.repositories[PurlType.DEB])
            ]
        ):
            return False

        config_deb_repository_urls = [str(repo.url) for repo in ChopprShares.config.deb_repositories]

        if set(deb_repository_urls) != set(config_deb_repository_urls):
            ChopprShares.log.error(
                "DEB repository mismatch: All repository URLs in the `manifest.yml` file should also be in the "
                "`transfer.yml` file"
            )

            for url in deb_repository_urls:
                ChopprShares.log.debug(f"Manifest Repository URL: {url}")
            for url in config_deb_repository_urls:
                ChopprShares.log.debug(f"Transfer Repository URL: {url}")

            return False

        repositories: list[apt_api.Repository] = []
        expected = sum(
            len(distribution.components)
            for repo in ChopprShares.config.deb_repositories
            for distribution in repo.distributions
        )
        ChopprShares.log.info(f"Pulling {expected} DEB repositories...")

        repositories_dir: Final[Path] = self.cache_dir / "repositories"
        repositories_dir.mkdir(parents=True, exist_ok=True)

        for repo in ChopprShares.config.deb_repositories:
            repo_url = parse_obj_as(HttpUrl, repo.url)
            for distribution in repo.distributions:
                for comp in distribution.components:
                    auth, verify = get_auth_and_verify(repo_url)
                    repositories.append(
                        apt_api.Repository(
                            repo_url,
                            distribution.name,
                            comp,
                            repositories_dir,
                            auth,
                            verify,
                        )
                    )

        self.repositories = apt_api.Sources(repositories)

        log_repo_pulls(expected, len(repositories), PurlType.DEB)

        return bool(repositories)

    def populate_potential_packages(self, file: str) -> None:
        """Get all potential packages that provide the given file.

        Arguments:
            file: Filename to search for
        """
        ChopprShares.log.debug(f"Getting Deb package associated with file: {file}")

        if self.repositories and (
            packages := {DebPackageData(package) for package in self.repositories.get_packages_that_provide(file)}
        ):
            for package in packages:
                ChopprShares.log.debug(f"Successfully found DEB: {package.name}")

                self.potential_packages.add(package)
                self.file_provided_by[file] = str(package)

    def populate_required_packages(self) -> None:
        """Get the packages that provide components in the SBOM."""
        for component in ChopprShares.components[PurlType.DEB]:
            if package := (
                ChopprShares.config.keep_essential_os_components and self._get_essential_package(component)
            ) or self._get_required_package(component):
                ChopprShares.log.debug(f"Package required: {package}")
                self.required_packages.add(package)

        ChopprShares.log.info(f"Found {len(self.required_packages)} required DEB packages")

        with self.cache_dir.joinpath("required_packages.txt").open("w") as output:
            output.writelines([f"{pkg}\n" for pkg in sorted(self.required_packages, key=lambda p: p.name)])

    def populate_all_nested_dependencies(self) -> None:
        """Get all nested dependenceis of the required packages."""
        for package in self.required_packages:
            ChopprShares.log.info(f"Populating nested DEB dependencies for {package.name}")
            amount_before = len(self.nested_dependencies)
            self._populate_nested_dependencies(package)
            ChopprShares.log.info(
                f"Found {len(self.nested_dependencies) - amount_before} new nested DEB dependencies for {package.name}"
            )

        ChopprShares.log.info(f"Found {len(self.nested_dependencies)} nested DEB dependencies")

        with self.cache_dir.joinpath("nested_dependencies.txt").open("w") as file:
            file.writelines([f"{pkg}\n" for pkg in sorted(self.nested_dependencies, key=lambda p: p.name)])

    ####################################################################################################
    # Utility Methods
    ####################################################################################################

    def _get_essential_package(self, component: Component) -> DebPackageData | None:
        if self.allow_version_mismatch:
            return next(
                (
                    DebPackageData(package)
                    for package in self.repositories.get_packages_by_name(component.name)
                    if _is_package_essential(package)
                ),
                None,
            )

        if component.version and (
            (package := self.repositories.get_package(component.name, component.version))
            and _is_package_essential(package)
        ):
            return DebPackageData(package)

        return None

    def _get_required_package(self, component: Component) -> DebPackageData | None:
        if self.allow_version_mismatch:
            return next(
                (package for package in self.potential_packages if str(package).startswith(component.name)),
                None,
            )

        if component.version:
            return next(
                (
                    package
                    for package in self.potential_packages
                    if str(package) == f"{component.name}-{component.version}"
                ),
                None,
            )

        return None

    def _get_new_packages_by_name(self, name: str) -> set[DebPackageData]:
        return {
            DebPackageData(p)
            for p in self.repositories.get_packages_by_name(name)
            if p not in self.required_packages and p not in self.nested_dependencies
        }

    def _evaluate_potential_nested_dependencies(
        self, potential_packages: set[DebPackageData], version: str, dependency_stack: set[str]
    ) -> bool:
        success = False
        for package in potential_packages:
            if DebHandler._is_package_version_valid(package, version):
                self.nested_dependencies.add(package)
                dependency_stack.add(package.name)
                self._populate_nested_dependencies(package, dependency_stack)
                success = True

        return success

    @limit_recursion()
    def _populate_nested_dependencies(self, data: DebPackageData, dependency_stack: set[str] | None = None) -> None:
        ChopprShares.log.debug(f"Dependencies for {data.name}: {', '.join([str(d) for d in data.package.depends])}")
        if not dependency_stack:
            dependency_stack = {data.name}

        for dependency in [d for d in data.package.depends if d.name not in dependency_stack]:
            if not self._evaluate_potential_nested_dependencies(
                self._get_new_packages_by_name(dependency.name), dependency.version, dependency_stack
            ):
                any(
                    self._evaluate_potential_nested_dependencies(
                        self._get_new_packages_by_name(alternate.name), alternate.version, dependency_stack
                    )
                    for alternate in dependency.alternates
                )

    @staticmethod
    def _get_max_version(versions: list[str]) -> tuple[bool, Version | None]:
        return next(
            (("=" in v, Version(v.replace("<", "").replace("=", "").strip())) for v in versions if "<" in v),
            (False, None),
        )

    @staticmethod
    def _get_min_version(versions: list[str]) -> tuple[bool, Version | None]:
        return next(
            (("=" in v, Version(v.replace(">", "").replace("=", "").strip())) for v in versions if ">" in v),
            (False, None),
        )

    @staticmethod
    def _package_version_within_max(package: DebPackageData, versions: list[str]) -> bool:
        can_be_equal, max_version = DebHandler._get_max_version(versions)

        if max_version is None:
            ChopprShares.log.debug(f"No valid maximum version: {', '.join(versions)}")
            return False

        return (can_be_equal and Version(package.version) <= max_version) or (
            not can_be_equal and Version(package.version) < max_version
        )

    @staticmethod
    def _package_version_within_min(package: DebPackageData, versions: list[str]) -> bool:
        can_be_equal, min_version = DebHandler._get_min_version(versions)

        if min_version is None:
            ChopprShares.log.debug(f"No valid minimum version: {', '.join(versions)}")
            return False

        return (can_be_equal and Version(package.version) >= min_version) or (
            not can_be_equal and Version(package.version) > min_version
        )

    @staticmethod
    def _package_version_within_range(package: DebPackageData, versions: list[str]) -> bool:
        within_max = DebHandler._package_version_within_max(package, versions)
        within_min = DebHandler._package_version_within_min(package, versions)

        return within_max and within_min

    @staticmethod
    def _is_package_version_valid(package: DebPackageData, version: str) -> bool:
        if version.startswith("="):
            exact_version = Version(version.replace("=", "").strip())
            return Version(package.version) == exact_version
        if all(sign in version for sign in ["<", ">"]):
            return DebHandler._package_version_within_range(package, version.split(","))
        if version.startswith("<"):
            return DebHandler._package_version_within_max(package, [version])
        if version.startswith(">"):
            return DebHandler._package_version_within_min(package, [version])

        return True



"""Dependency implementation."""

from __future__ import annotations

import re

from hashlib import sha256


__all__ = ["Dependency"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class Dependency:
    """Class that represents a Debian package dependency."""

    def __init__(self, dependencies: str) -> None:
        split_dependencies = [d.strip() for d in dependencies.split("|")]
        self.alternates: set[Dependency] = set()

        for idx, dependency in enumerate(split_dependencies):
            if idx == 0:
                dependency_match = re.match(r"(?P<name>.*)\s\((?P<version>.*)\)", dependency)
                self.name = str(dependency_match["name"]) if dependency_match else dependency
                self.version = str(dependency_match["version"]) if dependency_match else ""
            else:
                self.alternates.add(Dependency(dependency))

    def __str__(self) -> str:
        primary = f"{self.name}{' ' if self.version else ''}{self.version}"
        if self.alternates:
            return f"{primary} (Alternates: {', '.join(str(a) for a in self.alternates)})"
        return primary

    def __eq__(self, other: Dependency) -> bool:
        return self.name == other.name and self.version == other.version

    def __hash__(self) -> int:
        sha = sha256()
        sha.update(self.name.encode())
        sha.update(self.version.encode())
        return int(sha.hexdigest(), 16)



"""Sources implementation."""

from __future__ import annotations

from typing import TYPE_CHECKING

from hoppr import PurlType

from choppr.types import ChopprShares


if TYPE_CHECKING:
    from pydantic import HttpUrl

    from choppr.apt_api.binary_package import BinaryPackage
    from choppr.apt_api.repository import Repository


__all__ = ["Sources"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class Sources:
    """Class that represents a collection of APT repositories."""

    def __init__(self, repositories: list[Repository]) -> None:
        """Initialize an instance of APTSources.

        Arguments:
            repositories: List of Repository objects
        """
        self.__repositories = repositories
        sbom_package_names = [component.name for component in ChopprShares.components[PurlType.DEB]]
        self.packages = [
            package
            for repo in self.__repositories
            for package in repo.packages()
            if package.package in sbom_package_names
        ]

    def __getitem__(self, item: str) -> list[BinaryPackage]:
        """Get packages by name.

        Arguments:
            item: The package name

        Returns:
            list[BinaryPackage]: Packages with the given name
        """
        return self.get_packages_by_name(item)

    def get_package(self, name: str, version: str) -> BinaryPackage | None:
        """Return a single binary package.

        Arguments:
            name: Name of the package
            version: Version of the package

        Returns:
            BinaryPackage | None: The package if it exists
        """
        return next(
            (package for package in self.packages if package.package == name and package.version == version), None
        )

    def get_package_url(self, name: str, version: str) -> HttpUrl | None:
        """Return the URL of a single binary package.

        Arguments:
            name: Name of the package
            version: Version of the package

        Returns:
            HttpUrl | None: The package URL if it exists
        """
        for repo in self.__repositories:
            if url := repo.get_package_url(name, version):
                return url

        return None

    def get_packages_by_name(self, name: str) -> set[BinaryPackage]:
        """Return the list of available packages for a specific package name.

        Arguments:
            name: Name of the package

        Returns:
            set[BinaryPackage]: List of packages with the given name
        """
        return {package for package in self.packages if package.package == name}

    def get_packages_that_provide(self, filename: str) -> set[BinaryPackage]:
        """Determine what package, if any provides the given filename.

        Arguments:
            filename: The file to search for

        Returns:
            set[BinaryPackage]: The packages that provide the file
        """
        providers: set[BinaryPackage] = {
            package for package in self.packages for file in package.provides if filename.endswith(file)
        }
        return providers



"""Class definition for RpmPackageData."""

from __future__ import annotations

from hashlib import sha256
from typing import Any


__all__ = ["RpmPackageData"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class RpmPackageData:
    """Class containing frequently accessed information and the originating package.

    Members:
        - name
        - package
        - version
        - release_version
    """

    def __init__(self, package: dict[str, Any]) -> None:
        """Create an instance of PackageDetails.

        Arguments:
            package: The originating package dictionary
        """
        self.package = package
        self.name: str = package["name"]
        self.version: str = package["version"]["@ver"]
        self.release: str = package["version"]["@rel"]
        self.release_version = f"{self.version}-{self.release}"

    def __eq__(self, other: object) -> bool:
        match other:
            case RpmPackageData():
                return self.name == other.name and self.version == other.version
            case {"name": _, "version": {"@ver": _}}:
                return self.name == str(other["name"]) and self.version == str(other["version"]["@ver"])
            case _:
                return False

    def __hash__(self) -> int:
        sha = sha256()
        sha.update(self.name.encode())
        sha.update(self.version.encode())
        return int(sha.hexdigest(), 16)

    def __str__(self) -> str:
        return f"{self.name}-{self.release_version}"



"""Package defining custom types used within Choppr."""

from __future__ import annotations

from choppr.types.choppr_config import ChopprConfig, ChopprConfigModel
from choppr.types.choppr_shares import ChopprShares
from choppr.types.deb_package_data import DebPackageData
from choppr.types.rpm_package_data import RpmPackageData


__all__ = ["ChopprConfig", "ChopprConfigModel", "ChopprShares", "DebPackageData", "RpmPackageData"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"



"""Repository implementation."""

from __future__ import annotations

import json
import posixpath

from collections import defaultdict
from typing import TYPE_CHECKING

from pydantic import HttpUrl, parse_obj_as

from choppr.apt_api.packages_file import PackagesFile
from choppr.apt_api.release_file import ReleaseFile
from choppr.constants import DEFAULT_ARCH_DEB
from choppr.utils import HTTP, cache_file_outdated


if TYPE_CHECKING:
    from pathlib import Path

    from requests.auth import HTTPBasicAuth

    from choppr.apt_api.binary_package import BinaryPackage


__all__ = ["Repository"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class Repository:
    """Class that represents a single APT repository.

    Examples:
    ```python
    APTRepository("http://archive.ubuntu.com/ubuntu", "bionic", "main")
    APTRepository("https://pkg.jenkins.io/debian/", "binary")
    ```
    """

    repositories: list[str]

    def __init__(  # noqa: PLR0913, PLR0917
        self,
        url: HttpUrl,
        dist: str,
        component: str,
        cache_dir: Path,
        auth: HTTPBasicAuth | None = None,
        verify: str | bool = True,
        arch: str = DEFAULT_ARCH_DEB,
    ) -> None:
        """Initialize an instance of APTRepository.

        Arguments:
            url: The base URL of the repository
            dist: The target distribution
            component: The target component
            cache_dir: The cache directory to store cache files in
            auth: The credentials needed to access the URL (default None)
            verify: The certificate needed to access the URL (default True)
            arch: The CPU archetecture of the target (default DEFAULT_DEB_ARCH)
        """
        self.url = url
        self.auth = auth
        self.verify = verify
        self.dist = dist
        self.component = component

        self.repository_dir = cache_dir / f"{self.url.host}/dists/{self.dist}"
        self.repository_dir.mkdir(parents=True, exist_ok=True)

        self._cache_content_file = self.repository_dir / f"contents-{arch}.json"
        self._cache_release_file = self.repository_dir / "release"

        self._release_file: ReleaseFile | None = None
        self._packages: dict[str, list[BinaryPackage]] = {}

        if not self._cache_content_file.is_file() or cache_file_outdated(self._cache_content_file):
            obj = None
            content_file_url = parse_obj_as(
                HttpUrl,
                posixpath.join(self.url, "dists", self.dist, f"Contents-{arch}"),
            )

            if obj := HTTP.download_compressed(content_file_url, self.auth, self.verify):
                content_file_data = obj
            else:
                content_file_url = parse_obj_as(HttpUrl, posixpath.join(self.url, f"Contents-{arch}"))
                content_file_data = (
                    obj if (obj := HTTP.download_compressed(content_file_url, self.auth, self.verify)) else ""
                )

            file_list: dict[str, list[str]] = defaultdict(list[str])

            if content_file_data:
                for v, k in [
                    (file.split()[0], file.split()[1].split("/")[-1]) for file in content_file_data.strip().split("\n")
                ]:
                    file_list[k].append(v)

            with self._cache_content_file.open("w") as f:
                json.dump(file_list, f, indent=4, cls=json.JSONEncoder)

    def __getitem__(self, item: str) -> list[BinaryPackage]:
        """Get packages by name.

        Arguments:
            item: The package name

        Returns:
            list[BinaryPackage]: Packages with the given name
        """
        return self.get_packages_by_name(item)

    @property
    def all_components(self) -> list[str]:
        """Returns the all components of this repository.

        Returns:
            list[str]: List of components in the repository
        """
        return self.release_file.components if self.release_file else []

    @property
    def release_file(self) -> ReleaseFile | None:
        """Returns the Release file of this repository.

        Returns:
            ReleaseFile | None: The release file if it exists
        """
        if not self._cache_release_file.is_file() or cache_file_outdated(self._cache_release_file):
            url = parse_obj_as(HttpUrl, posixpath.join(self.url, "dists", self.dist, "Release"))
            if release_content := HTTP.download(url, self.auth, self.verify):
                with self._cache_release_file.open("w") as f:
                    f.write(release_content)
                self._release_file = ReleaseFile(release_content)
        elif not self._release_file:
            with self._cache_release_file.open() as f:
                self._release_file = ReleaseFile(f.read())

        return self._release_file

    def packages(self, arch: str = DEFAULT_ARCH_DEB) -> list[BinaryPackage]:
        """Return all binary packages of this repository.

        Arguments:
            arch: The architecture to return packages for (default DEFAULT_DEB_ARCH)

        Returns:
            list[BinaryPackage]: List of BinaryPackages for given architecture
        """
        if not self._packages.get(arch):
            cache_package_file = self.repository_dir / f"{self.component}-packages-{arch}"

            if not cache_package_file.is_file() or cache_file_outdated(cache_package_file):
                package_file_url = parse_obj_as(
                    HttpUrl,
                    posixpath.join(
                        self.url,
                        "dists",
                        self.dist,
                        self.component,
                        f"binary-{arch}",
                        "Packages",
                    ),
                )

                if package_file_content := HTTP.download_compressed(package_file_url, self.auth, self.verify):
                    with cache_package_file.open("w", encoding="utf-16") as file:
                        file.write(package_file_content.strip())

            self._packages[arch] = PackagesFile(cache_package_file, self._cache_content_file, self.url).packages

        return self._packages[arch]

    def get_package(self, name: str, version: str) -> BinaryPackage | None:
        """Return a single binary package.

        Arguments:
            name: Name of the package
            version: Version of the package

        Returns:
            BinaryPackage | None: Package with the given name and version
        """
        return next(
            (package for package in self.packages() if package.package == name and package.version == version),
            None,
        )

    def get_package_url(self, name: str, version: str) -> HttpUrl | None:
        """Return the URL for a single binary package.

        Arguments:
            name: Name of the package
            version: Version of the package

        Returns:
            HttpUrl | None: URL for package
        """
        if package := self.get_package(name, version):
            url: HttpUrl = parse_obj_as(HttpUrl, posixpath.join(self.url, package.filename))
            return url

        return None

    def get_packages_by_name(self, name: str) -> list[BinaryPackage]:
        """Return the list of available packages for a specific package name.

        Arguments:
            name: Name of the package

        Returns:
            list[BinaryPackage]: List of packages that match the given name
        """
        return [package for package in self.packages() if package.package == name]

    def get_common_packages(self, packages: list[str]) -> list[str]:
        """Return the list of packages that are in the repository and given list.

        Arguments:
            packages: String list of packages in SBOM

        Returns:
            list[str]: List of packages that are in common with the provided list
        """
        return list(set(packages) & {package.package for package in self.packages()})



"""Module with decorator functions."""

from __future__ import annotations

import functools

from typing import TYPE_CHECKING, Any, TypeVar

from choppr.types import ChopprShares


if TYPE_CHECKING:
    from collections.abc import Callable


__all__ = ["limit_recursion"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"

T = TypeVar("T")


def limit_recursion(max_depth: int | None = None) -> Callable[[Callable[..., T]], Callable[..., T | None]]:
    """Limit the recursion depth of a function.

    Returns None instead of the function result if the limit is reached.

    Arguments:
        max_depth: The maximum allowed recursion depth (default None)

    Returns:
        A decorator function that wraps the original function
    """

    def decorator(func: Callable[..., T]) -> Callable[..., T | None]:
        depth = 0

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> T | None:  # noqa: ANN401
            nonlocal max_depth
            if max_depth is None:
                max_depth = ChopprShares.config.recursion_limit

            nonlocal depth
            depth += 1
            result = None if depth > max_depth else func(*args, **kwargs)
            depth -= 1
            return result

        return wrapper

    return decorator



"""Class definition for ChopprConfig."""

from __future__ import annotations

import re

from datetime import timedelta
from hashlib import sha256
from pathlib import Path
from typing import Any

from hoppr import Component, PurlType
from pydantic import (  # noqa: TC002
    BaseModel,
    Field,
    HttpUrl,
    PositiveFloat,
    PositiveInt,
    PrivateAttr,
    parse_obj_as,
    validator,
)

from choppr import strace
from choppr.constants import COMPONENT_LIST_FORMATS, DEFAULT_RECURSION_LIMIT


__all__ = ["ChopprConfig", "ChopprConfigModel", "DebianDistribution", "DebianRepository"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


def _validate_file_path(value: str) -> Path:
    path = Path(value)

    if not path.is_file():
        raise ValueError(f"File does not exist: {path}")
    return path


def _validate_http_url(value: str) -> HttpUrl:
    return parse_obj_as(HttpUrl, value.rstrip("/"))


def _validate_regex(value: str) -> re.Pattern:
    try:
        return re.compile(value)
    except re.error as e:
        raise ValueError(f"Invalid regular expression: {value}") from e


def _default_excluded_components() -> dict[PurlType, ExcludedComponentsFile]:
    return {
        purl_type: ExcludedComponentsFile(
            file=Path(f"choppr_excluded_components_{purl_type.name.lower()}.txt"),
            component_format=COMPONENT_LIST_FORMATS.get(purl_type, "{name}={version}"),
        )
        for purl_type in PurlType
    }


class Certificate(BaseModel):
    # Required Attributes
    url: str
    certificate: Path

    # Validators
    _validate_certificate = validator("certificate", pre=True, allow_reuse=True)(_validate_file_path)


class DebianDistribution(BaseModel):
    """Class representation for a debian distribution, to include its name and components.

    Members:
        - name
        - components
    """

    name: str
    components: list[str]


class DebianRepository(BaseModel):
    """Class representation for a debian repository, to include its URL and distributions.

    Members:
        - url
        - distributions
    """

    url: HttpUrl
    distributions: list[DebianDistribution]

    _validate_url = validator("url", pre=True, allow_reuse=True)(_validate_http_url)


class HttpRequestLimits(BaseModel):
    """Class with values to configure HTTP request limits.

    Members:
        - retries
        - retry_interval
        - timeout
    """

    retries: PositiveInt = 3
    retry_interval: PositiveFloat = 5.0
    timeout: PositiveFloat = 60.0


class ExcludedComponentsFile(BaseModel):
    """The filename to output excluded components to, and what format to write them as.

    Members:
        - file
        - format
    """

    file: Path
    component_format: str = ""


class OutputFiles(BaseModel):
    """Class with values for the output files for Choppr.

    Members:
        - excluded_components
    """

    excluded_components: dict[PurlType, ExcludedComponentsFile] = Field(default_factory=_default_excluded_components)

    @validator("excluded_components", pre=True, allow_reuse=True)
    @classmethod
    def _validate_excluded_components(cls, value: dict[str, dict[str, str]]) -> dict[PurlType, ExcludedComponentsFile]:
        excluded_components: dict[PurlType, ExcludedComponentsFile] = _default_excluded_components()
        try:
            for purl, file_and_format in value.items():
                purl_type = PurlType[purl.upper()]
                excluded_components_file = ExcludedComponentsFile(**file_and_format)
                excluded_components[purl_type].file = excluded_components_file.file
                if excluded_components_file.component_format:
                    excluded_components[purl_type].component_format = excluded_components_file.component_format
        except KeyError as e:
            raise ValueError(
                f"Invalid purl type: {e} - Accpeted values: [{', '.join(m.name for m in PurlType)}]"
            ) from e
        else:
            return excluded_components


class PackagePattern(BaseModel):
    """Class with the name, version, and purl type for a package.

    Members:
        - name
        - version
    """

    name: re.Pattern
    version: re.Pattern

    _validate_patterns = validator("name", "version", pre=True, allow_reuse=True)(_validate_regex)

    def __eq__(self, other: PackagePattern | Component) -> bool:
        if isinstance(other, PackagePattern):
            return self.name == other.name and self.version == other.version
        if isinstance(other, Component):
            return self.name.match(other.name) and self.version.match(other.version)
        return False

    def __hash__(self) -> int:
        sha = sha256()
        sha.update(str(self.name).encode())
        sha.update(str(self.version).encode())
        return int(sha.hexdigest(), 16)


class ChopprConfigModel(BaseModel):
    """Class to validate and parse the configuration values provided to ChopprPlugin.

    Members:
        - strace_results
        - allow_version_mismatch
        - allowlist
        - cache_dir
        - cache_timeout
        - certificates
        - clear_cache
        - deb_repositories
        - delete_excluded
        - denylist
        - http_limits
        - keep_essential_os_components
        - output_files
        - recursion_limit
        - strace_regex_excludes

    Methods:
        - strace_files
    """

    HttpRequestLimits.update_forward_refs()

    # Required Attributes
    strace_results: Path
    # Optional Attributes
    allow_version_mismatch: bool = False
    allowlist: dict[PurlType, set[PackagePattern]] = Field(default={})
    cache_dir: Path = Field(default_factory=lambda: Path.cwd() / ".cache" / "choppr")
    cache_timeout: timedelta = Field(default=timedelta(days=7))
    certificates: dict[str, Path] = Field(default={})
    clear_cache: bool = False
    deb_repositories: list[DebianRepository] = Field(default=[])
    delete_excluded: bool = True
    denylist: dict[PurlType, set[PackagePattern]] = Field(default={})
    http_limits: HttpRequestLimits = Field(default=HttpRequestLimits())
    keep_essential_os_components: bool = False
    output_files: OutputFiles = Field(default=OutputFiles())
    recursion_limit: PositiveInt = DEFAULT_RECURSION_LIMIT
    strace_regex_excludes: list[re.Pattern] = Field(default=[])
    # Private Attributes
    _strace_files: set[str] = PrivateAttr(default=set())

    # Validators
    _validate_strace_results = validator("strace_results", allow_reuse=True)(_validate_file_path)
    _validate_strace_regex_excludes = validator("strace_regex_excludes", pre=True, each_item=True, allow_reuse=True)(
        _validate_regex
    )

    @validator("cache_timeout", pre=True, allow_reuse=True)
    @classmethod
    def _validate_cache_timeout(cls, cache_timeout: str) -> timedelta:
        timedelta_pattern = re.compile(r"^(?P<duration>\d+)\s?(?P<unit>d|h|m|s)$", re.IGNORECASE)

        timeout_match = timedelta_pattern.match(cache_timeout)

        error_message = "Invalid 'cache_timeout' value: Expected a number followed by a unit (d, h, m, s)"

        if not timeout_match:
            raise ValueError(error_message)

        timeout_duration = int(timeout_match["duration"])
        timeout_unit = timeout_match["unit"].lower()

        unit_map = {
            "d": "days",
            "h": "hours",
            "m": "minutes",
            "s": "seconds",
        }

        if timeout_unit not in unit_map:
            raise ValueError(error_message)

        return timedelta(**{unit_map[timeout_unit]: timeout_duration})

    @validator("certificates", pre=True, allow_reuse=True)
    @classmethod
    def _validate_certificates(cls, certificates: list[dict[str, str]]) -> dict[str, Path]:
        certificate_map: dict[str, Path] = {}

        for certificate in certificates:
            c = Certificate(**certificate)
            certificate_map[c.url] = c.certificate

        return certificate_map

    @validator("allowlist", "denylist", pre=True, allow_reuse=True)
    @classmethod
    def _validate_exception_list(cls, value: dict[str, set[PackagePattern]]) -> dict[PurlType, set[PackagePattern]]:
        try:
            return {PurlType[purl_type.upper()]: packages for purl_type, packages in value.items()}
        except KeyError as e:
            raise ValueError(
                f"Invalid purl type: {e} - Accpeted values: [{', '.join(m.name for m in PurlType)}]"
            ) from e

    @validator("denylist", allow_reuse=True)
    @classmethod
    def _validate_exception_overlap(
        cls, denylist: dict[PurlType, set[PackagePattern]], values: dict[str, Any]
    ) -> dict[PurlType, set[PackagePattern]]:
        allowlist: dict[PurlType, set[PackagePattern]] = values.get("allowlist")

        for purl_type in PurlType:
            if allowlist.get(purl_type, set()) & denylist.get(purl_type, set()):
                raise ValueError(f"The allowlist and denylist have at least one overlapping {purl_type.name} package")

        return denylist

    def strace_files(self) -> set[str]:
        """List of files parsed from the provided strace results.

        Returns:
            set[str]: List of files found from strace
        """
        if not self._strace_files:
            parsed_strace_files = strace.get_files(self.strace_results)

            if self.strace_regex_excludes:
                self._strace_files = {
                    file
                    for file in parsed_strace_files
                    if not any(bool(re.search(exclude, file)) for exclude in self.strace_regex_excludes)
                }
                self.cache_dir.mkdir(parents=True, exist_ok=True)
                with self.cache_dir.joinpath("filtered_strace_results.txt").open("w") as output:
                    output.writelines([f"{file}\n" for file in self._strace_files])
            else:
                self._strace_files = parsed_strace_files

        return self._strace_files


class ChopprConfig:
    """A class to store the Choppr configuration.

    Members:
        - allow_version_mismatch
        - allowlist
        - cache_dir
        - cache_timeout
        - certificates
        - clear_cache
        - deb_repositories
        - delete_excluded
        - denylist
        - http_limits
        - keep_essential_os_components
        - output_files
        - recursion_limit
        - strace_files
    """

    def __init__(self, model: ChopprConfigModel) -> None:
        self.allow_version_mismatch: bool = model.allow_version_mismatch
        self.allowlist: dict[PurlType, list[PackagePattern]] = model.allowlist
        self.cache_dir: Path = model.cache_dir
        self.cache_timeout: timedelta = model.cache_timeout
        self.certificates: dict[str, Path] = model.certificates
        self.clear_cache: bool = model.clear_cache
        self.deb_repositories: list[DebianRepository] = model.deb_repositories
        self.delete_excluded: bool = model.delete_excluded
        self.denylist: dict[PurlType, list[PackagePattern]] = model.denylist
        self.http_limits: HttpRequestLimits = model.http_limits
        self.keep_essential_os_components: bool = model.keep_essential_os_components
        self.output_files: OutputFiles = model.output_files
        self.recursion_limit: PositiveInt = model.recursion_limit
        self.strace_files: set[str] = model.strace_files()



"""BinaryPackage implementation."""

from __future__ import annotations

import io
import os
import tarfile

from typing import Any

import zstandard as zstd

from pydantic import HttpUrl, parse_obj_as

from choppr.apt_api._utils import get_list, get_value
from choppr.apt_api.dependency import Dependency
from choppr.constants import DEB_HEADER_LENGTH


__all__ = ["BinaryPackage"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


def _flatten(matrix: list[Any]) -> list[str]:
    flat_list = []
    for element in matrix:
        if isinstance(element, list):
            flat_list.extend(_flatten(element))
        else:
            flat_list.append(element.strip())
    return flat_list


def _extract_archive(archive_file: bytes) -> dict[str, bytes]:
    archive = {}

    with io.BytesIO(archive_file) as f:
        # Read the ar file header (archive starts with "!<arch>\n")
        if f.read(8) != b"!<arch>\n":
            # Invalid ar archive
            return {}

        # Read the file metadata in the archive
        while (header := f.read(DEB_HEADER_LENGTH)) and len(header) >= DEB_HEADER_LENGTH:
            file_name = header[:16].strip().decode()
            file_size = int(header[48:58].strip())

            # Read the file content
            content = f.read(file_size)

            # Skip over any padding
            # (ar files are padded to even byte boundaries)
            if file_size % 2 != 0:
                f.seek(1, os.SEEK_CUR)

            # Write the file contents to the destination folder
            archive[file_name] = content

    return archive


def _get_deb_files(deb_file: bytes) -> list[str]:
    # Step 1: Extract the ar archive from the .deb file
    archive = _extract_archive(deb_file)

    # Step 2: Find the data.tar.gz or data.tar.xz in the extracted contents
    data_archive = None
    for filename in archive:
        if filename.startswith("data.tar"):
            data_archive = archive[filename]
            break

    # Step 3: Extract the data archive (the package contents)
    if data_archive:
        if ".zst" in filename:
            with io.BytesIO(data_archive) as f:
                dctx = zstd.ZstdDecompressor()
                with dctx.stream_reader(f) as reader:
                    buffer = io.BytesIO(reader.read())
                    buffer.seek(0)
                    fileobj = buffer
        else:
            fileobj = io.BytesIO(data_archive)

        # Extract .tar archive (either data.tar.xz or data.tar.gz)
        with tarfile.open(fileobj=fileobj) as tar:
            members = tar.getmembers()
            return [member.name[1:] for member in members if member.isfile()]
    else:
        return []


class BinaryPackage:  # noqa: PLR0904
    """Class that represents a binary Debian package."""

    def __init__(self, content: str, file_list: list[str] | None = None, package_url: str = "") -> None:
        """Initialize an instance of BinaryPackage.

        Arguments:
            content: The section of the Packages file for this specific package
            file_list: The list of files internal to this specific package (default None)
            package_url: The url of the package (default "")
        """
        if file_list is None:
            file_list = []
        self.__content = content.strip()
        self.file_list: list[str] = file_list
        self.package_url = parse_obj_as(HttpUrl, f"{package_url}/{get_value(self.__content, 'Filename')}")

        pre_depends: set[Dependency] = {Dependency(d) for d in get_list(self.__content, "Pre-Depends")}
        depends: set[Dependency] = {Dependency(d) for d in get_list(self.__content, "Depends")}
        self.depends = pre_depends.union(depends)

    @property
    def package(self) -> str:
        """Get the value of package.

        Returns:
            str: Package value
        """
        return str(get_value(self.__content, "Package"))

    @property
    def version(self) -> str:
        """Get the value of version.

        Returns:
            str: Version value
        """
        return str(get_value(self.__content, "Version"))

    @property
    def filename(self) -> str:
        """Get the value of filename.

        Returns:
            str: Filename value
        """
        return str(get_value(self.__content, "Filename"))

    @property
    def maintainer(self) -> str | None:
        """Get the value of maintainer.

        Returns:
            str: Maintainer value
        """
        return get_value(self.__content, "Maintainer")

    @property
    def original_maintainer(self) -> str | None:
        """Get the value of Original-Maintainer.

        Returns:
            str | None: Original-Maintainer value if it exists
        """
        return get_value(self.__content, "Original-Maintainer")

    @property
    def architecture(self) -> str | None:
        """Get the value of architecture.

        Returns:
            str | None: Architecture value if it exists
        """
        return get_value(self.__content, "Architecture")

    @property
    def multi_arch(self) -> str | None:
        """Get the value of Multi-Arch.

        Returns:
            str | None: Multi-Arch value if it exists
        """
        return get_value(self.__content, "Multi-Arch")

    @property
    def homepage(self) -> str | None:
        """Get the value of Homepage.

        Returns:
            str | None: Homepage value if it exists
        """
        return get_value(self.__content, "Homepage")

    @property
    def origin(self) -> str | None:
        """Get the value of Origin.

        Returns:
            str | None: Origin value if it exists
        """
        return get_value(self.__content, "Origin")

    @property
    def priority(self) -> str | None:
        """Get the value of Priority.

        Returns:
            str | None: Priority value if it exists
        """
        return get_value(self.__content, "Priority")

    @property
    def essential(self) -> bool | None:
        """Get the value of Essential.

        Returns:
            bool | None: Essential value if it exists
        """
        return bool((value := get_value(self.__content, "Essential")) and value.lower() == "yes")

    @property
    def build_essential(self) -> bool | None:
        """Get the value of Build-Essential.

        Returns:
            bool | None: Build-Essential value if it exists
        """
        return bool((value := get_value(self.__content, "Build-Essential")) and value.lower() == "yes")

    @property
    def section(self) -> str | None:
        """Get the value of Section.

        Returns:
            str | None: Section value if it exists
        """
        return get_value(self.__content, "Section")

    @property
    def provides(self) -> list[str]:
        """Get the value of Provides.

        Returns:
            str | None: Provides value if it exists
        """
        from choppr.utils import HTTP  # noqa: PLC0415

        if self.file_list:
            return self.file_list
        if deb := HTTP.download_raw(self.package_url):
            self.file_list = _get_deb_files(deb)
        return self.file_list

    @property
    def replaces(self) -> list[str]:
        """Get the value of Replaces.

        Returns:
            str | None: Replaces value if it exists
        """
        return get_list(self.__content, "Replaces")

    @property
    def breaks(self) -> list[str]:
        """Get the value of Breaks.

        Returns:
            str | None: Breaks value if it exists
        """
        return get_list(self.__content, "Breaks")

    @property
    def recommends(self) -> list[str]:
        """Get the value of Recommends.

        Returns:
            str | None: Recommends value if it exists
        """
        return get_list(self.__content, "Recommends")

    @property
    def suggests(self) -> list[str]:
        """Get the value of Suggests.

        Returns:
            str | None: Suggests value if it exists
        """
        return get_list(self.__content, "Suggests")

    @property
    def conflicts(self) -> list[str]:
        """Get the value of Conflicts.

        Returns:
            str | None: Conflicts value if it exists
        """
        return get_list(self.__content, "Conflicts")

    @property
    def installed_size(self) -> str | None:
        """Get the value of Installed-Size.

        Returns:
            str | None: Installed-Size value if it exists
        """
        return get_value(self.__content, "Installed-Size")

    @property
    def size(self) -> str | None:
        """Get the value of Size.

        Returns:
            str | None: Size value if it exists
        """
        return get_value(self.__content, "Size")

    @property
    def md5(self) -> str | None:
        """Get the value of MD5Sum.

        Returns:
            str | None: MD5Sum value if it exists
        """
        return get_value(self.__content, "MD5Sum")

    @property
    def sha1(self) -> str | None:
        """Get the value of SHA1.

        Returns:
            str | None: SHA1 value if it exists
        """
        return get_value(self.__content, "SHA1")

    @property
    def sha256(self) -> str | None:
        """Get the value of SHA256.

        Returns:
            str | None: SHA256 value if it exists
        """
        return get_value(self.__content, "SHA256")

    @property
    def description(self) -> str | None:
        """Get the value of Description.

        Returns:
            str | None: Description value if it exists
        """
        return get_value(self.__content, "Description")

    @property
    def description_md5(self) -> str | None:
        """Get the value of Description-md5.

        Returns:
            str | None: Description-md5 value if it exists
        """
        return get_value(self.__content, "Description-md5")

    @property
    def built_using(self) -> str | None:
        """Get the value of Built-Using.

        Returns:
            str | None: Built-Using value if it exists
        """
        return get_value(self.__content, "Built-Using")

    @property
    def source(self) -> str | None:
        """Get the value of Source.

        Returns:
            str | None: Source value if it exists
        """
        return get_value(self.__content, "Source")

    @property
    def task(self) -> str | None:
        """Get the value of Task.

        Returns:
            str | None: Task value if it exists
        """
        return get_value(self.__content, "Task")

    @property
    def supported(self) -> str | None:
        """Get the value of Supported.

        Returns:
            str | None: Supported value if it exists
        """
        return get_value(self.__content, "Supported")



"""Package defining an interface to handle apt packages."""

from __future__ import annotations

from choppr.apt_api.binary_package import BinaryPackage
from choppr.apt_api.dependency import Dependency
from choppr.apt_api.packages_file import PackagesFile
from choppr.apt_api.release_file import ReleaseFile
from choppr.apt_api.repository import Repository
from choppr.apt_api.sources import Sources


__all__ = ["BinaryPackage", "Dependency", "PackagesFile", "ReleaseFile", "Repository", "Sources"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"



"""Module for handling RPM packages."""

from __future__ import annotations

import gzip
import json
import lzma

from copy import deepcopy
from http import HTTPStatus
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final, cast

import jmespath
import xmltodict

from hoppr.models.manifest import Repository
from hoppr.models.types import PurlType
from pydantic import HttpUrl, parse_obj_as

from choppr.constants import DEFAULT_ARCH_RPM
from choppr.decorators import limit_recursion
from choppr.types import ChopprShares, RpmPackageData
from choppr.utils import HTTP, get_auth_and_verify, log_repo_pulls, remove_parenthesis


if TYPE_CHECKING:
    from collections import OrderedDict

    from hoppr import Component


__all__ = ["RpmHandler"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class RpmHandler:
    """Class to handle all RPM repository processing."""

    def __init__(self) -> None:
        self.allow_version_mismatch = ChopprShares.config.allow_version_mismatch
        self.cache_dir = ChopprShares.config.cache_dir / "rpm"
        self.cache_dir.mkdir(parents=True, exist_ok=True)

        self.file_provided_by: dict[str, str] = {}
        self.repositories: dict[str, OrderedDict[str, Any]] = {}
        self.potential_packages: set[RpmPackageData] = set()
        self.required_packages: set[RpmPackageData] = set()
        self.nested_dependencies: set[RpmPackageData] = set()

    ####################################################################################################
    # Exported Methods
    ####################################################################################################

    def populate_repositories(self) -> bool:
        """Pull all of the metadata for RPM repositories provided in the config.

        Returns:
            bool: True when any repository was successfully pulled
        """
        if rpm_repository_urls := [
            parse_obj_as(HttpUrl, str(repo.url))
            for repo in parse_obj_as(list[Repository], ChopprShares.context.repositories[PurlType.RPM])
        ]:
            expected = len(rpm_repository_urls)
            ChopprShares.log.info(f"Pulling {expected} RPM repositories...")

            for repo_url in rpm_repository_urls:
                if repo_return := self._get_remote_repo(repo_url):
                    repo_checksum, repo_data = repo_return
                    self.repositories[repo_checksum] = repo_data
            log_repo_pulls(expected, len(self.repositories), PurlType.RPM)

        return bool(len(self.repositories))

    def populate_potential_packages(self, file: str, search_files: bool) -> None:
        """Get all potential packages that provide the given file.

        Arguments:
            file: Filename to search for
            search_files: Enable searching the files section of RPM repo metadata
        """
        ChopprShares.log.debug(f"Getting RPM associated with file: {file}")

        if packages := self._rpm_provides(file, search_files):
            for package in packages:
                ChopprShares.log.debug(f"Successfully found RPM: {package.name}")

                self.potential_packages.add(package)
                self.file_provided_by[file] = str(package)

    def populate_required_packages(self) -> None:
        """Get the packages that provide components in the SBOM."""
        for component in ChopprShares.components[PurlType.RPM]:
            if package := self._get_component_package(component):
                ChopprShares.log.debug(f"Package required: {package}")
                self.required_packages.add(package)

        ChopprShares.log.info(f"Found {len(self.required_packages)} required rpm packages")

        with self.cache_dir.joinpath("required_packages.txt").open("w") as output:
            output.writelines([f"{pkg}\n" for pkg in sorted(self.required_packages, key=lambda p: p.name)])

    def populate_all_nested_dependencies(self) -> None:
        """Get all nested dependenceis of the required packages."""
        for package in deepcopy(self.required_packages):
            if package_requires := self._rpm_requires(package):
                ChopprShares.log.debug(f"Getting nested rpm dependencies for {package}")
                amount_before = len(self.nested_dependencies)
                self._populate_nested_dependencies(package_requires)
                ChopprShares.log.debug(
                    f"Found {len(self.nested_dependencies) - amount_before} new nested rpm dependencies"
                )

        ChopprShares.log.info(f"Found {len(self.nested_dependencies)} nested rpm dependencies")

        with self.cache_dir.joinpath("nested_dependencies.txt").open("w") as file:
            file.writelines([f"{pkg}\n" for pkg in sorted(self.nested_dependencies, key=lambda p: p.name)])

    ####################################################################################################
    # Utility Methods
    ####################################################################################################

    def _get_remote_repo(self, repo_url: HttpUrl) -> tuple[str, OrderedDict[str, Any]] | None:
        repositories_dir: Final[Path] = self.cache_dir.joinpath("repositories", repo_url.host)
        repositories_dir.mkdir(parents=True, exist_ok=True)

        url = parse_obj_as(HttpUrl, f"{repo_url}/repodata/repomd.xml")
        ChopprShares.log.info(f"Pulling repository data from {url}...")

        basic_auth, verify = get_auth_and_verify(url)

        with HTTP.get(url, basic_auth, verify) as response:
            if response.status_code > HTTPStatus.MULTIPLE_CHOICES:
                ChopprShares.log.error(f"Failed to pull repository: {response.content.decode()}")
                return None

            repomd_dict = xmltodict.parse(response.content)

            primary_checksum = jmespath.search(
                expression='[?"@type"==\'primary\'].checksum."#text"',
                data=repomd_dict["repomd"]["data"],
            )
            primary_location = jmespath.search(
                expression='[?"@type"==\'primary\'].location."@href"',
                data=repomd_dict["repomd"]["data"],
            )

            if isinstance(primary_checksum, list) and len(primary_checksum) == 1:
                cache_repository_file = repositories_dir / f"{primary_checksum[0]}.json"

                # Use cached repository
                if cache_repository_file.is_file():
                    with cache_repository_file.open() as file:
                        ChopprShares.log.info("Successfully loaded repository from cache")
                        return (primary_checksum[0], cast("OrderedDict[str, Any]", json.load(file)))
                # Pull repository
                elif isinstance(primary_location, list) and len(primary_location) == 1:
                    primary_url = parse_obj_as(HttpUrl, f"{repo_url}/{primary_location[0]}")
                    with HTTP.get(primary_url, basic_auth, verify) as primary_response:
                        if primary_response.status_code > HTTPStatus.MULTIPLE_CHOICES:
                            ChopprShares.log.error(f"Failed to pull repository: {response.content.decode()}")
                            return None

                        match primary_url:
                            case _ if primary_url.endswith(".gz"):
                                primary_xml = gzip.decompress(primary_response.content)
                            case _ if primary_url.endswith(".xz"):
                                primary_xml = lzma.decompress(primary_response.content)
                            case _ if primary_url.endswith(".xml"):
                                primary_xml = primary_response.content
                            case _:
                                ChopprShares.log.error(
                                    f"Unsupported file type found for primary repository: {primary_url}"
                                )
                                return None

                        ChopprShares.log.info("Successfully pulled repository data")

                        repo_data = xmltodict.parse(primary_xml)
                        # Write repository data to cache
                        with cache_repository_file.open("w") as file:
                            json.dump(repo_data, file, indent=4)
                        return (primary_checksum[0], repo_data)

        ChopprShares.log.error("Failed to get `location` and/or `checksum` from repository")
        return None

    def _rpm_provides(
        self, filename: str, search_files: bool = False, arch: str = DEFAULT_ARCH_RPM
    ) -> set[RpmPackageData]:
        if filename in self.file_provided_by:
            return set()

        providers: set[RpmPackageData] = set()

        for repo in self.repositories.values():
            package_data: set[RpmPackageData] = {
                RpmPackageData(package)
                for package in jmespath.search(
                    expression=f"metadata.package[?arch == '{arch}' || arch == 'noarch']", data=repo
                )
            }
            if search_files:
                providers.update(self._search_files_section(filename, package_data))

            if not providers:
                providers.update(self._search_rpm_provides_section(filename, package_data))

        return providers

    def _get_component_package(self, component: Component) -> RpmPackageData | None:
        if not self.allow_version_mismatch:
            if not component.version:
                return None
            return next(
                (
                    package
                    for package in self.potential_packages
                    if str(package) == f"{component.name}-{component.version}"
                ),
                None,
            )
        return next(
            (package for package in self.potential_packages if str(package).startswith(component.name)),
            None,
        )

    @limit_recursion()
    def _populate_nested_dependencies(self, required_files: set[str]) -> None:
        for file in required_files:
            if packages := {
                package
                for package in self._rpm_provides(file)
                if package not in self.required_packages | self.nested_dependencies
            }:
                for package in packages:
                    self.nested_dependencies.add(package)

                    if package_requires := self._rpm_requires(package):
                        self._populate_nested_dependencies(package_requires)

    @staticmethod
    def _rpm_requires(data: RpmPackageData) -> set[str] | None:
        if requirements := jmespath.search(
            expression='format."rpm:requires"."rpm:entry"[*]."@name"',
            data=data.package,
        ):
            return set(requirements)

        return None

    @staticmethod
    def _search_files_section(file_path: str, package_data: set[RpmPackageData]) -> set[RpmPackageData]:
        """Search the files secetion of a given package for a given file.

        It will search files for an exact match, and for directories, it will check if the given
        file path starts with the directory path.

        Arguments:
            file_path: The file to search for
            package_data: The package to search

        Returns:
            set[RpmPackageData]: Packages containing the file
        """
        # Search files
        providers: set[RpmPackageData] = {
            data
            for data in package_data
            if jmespath.search(
                expression=(f"format.file[?type(@) == 'string' && @ == '{file_path}']"),
                data=data.package,
            )
        }
        if not providers:
            # Search directories
            for data in package_data:
                directories: list[str] | None = jmespath.search(
                    expression='format."file"[?"@type" == \'dir\']."#text"',
                    data=data.package,
                )
                if directories and any(file_path.startswith(directory) for directory in directories):
                    providers.add(data)

        return providers

    @staticmethod
    def _search_rpm_provides_section(file_path: str, package_data: set[RpmPackageData]) -> set[RpmPackageData]:
        """Search the rpm:provides secetion of a given package for a given file.

        Arguments:
            file_path: The file to search for
            package_data: The package to search

        Returns:
            set[RpmPackageData]: Packages containing the file
        """
        providers: set[RpmPackageData] = set()

        for data in package_data:
            # Search provides section
            package_provides: list[str] | None = jmespath.search(
                expression=('format."rpm:provides"."rpm:entry"[*]."@name"'),
                data=data.package,
            )
            if package_provides and any(
                remove_parenthesis(provides) == Path(file_path).name for provides in package_provides
            ):
                providers.add(data)

        return providers



"""Utility functions used within the Choppr plugin.

Functions:
    - append_slash
    - http_get
    - remove_parenthesis

Classes:
    - RPM
"""

from __future__ import annotations

import bz2
import contextlib
import gzip
import lzma
import re
import time

from datetime import datetime
from http.client import RemoteDisconnected
from typing import TYPE_CHECKING

import requests

from hoppr import Credentials, PurlType
from pydantic import HttpUrl, SecretStr, parse_obj_as
from requests.auth import HTTPBasicAuth
from requests.exceptions import ProxyError

from choppr.constants import LOG_HEADER_LENGTH
from choppr.types import ChopprShares


if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from hoppr import Component
    from requests import Response


__all__ = [
    "HTTP",
    "cache_file_outdated",
    "get_auth_and_verify",
    "get_component_type",
    "log_header",
    "log_repo_pulls",
    "output_list",
    "prepend_slash",
    "remove_parenthesis",
]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class HTTP:
    """Utility methods to handle HTTP requests."""

    @staticmethod
    def get(url: HttpUrl, auth: HTTPBasicAuth | None = None, verify: str | bool = True) -> Response:
        """Perform an HTTP get request with the provided parameters.

        Arguments:
            url: The URL to access
            auth: The credentials needed to access the URL (default None)
            verify: The certificate needed to access the URL (default True)

        Returns:
            Response: The response from the request
        """
        limits = ChopprShares.config.http_limits

        for attempt in range(limits.retries):
            if attempt > 0:
                time.sleep(limits.retry_interval)

            with contextlib.suppress(RemoteDisconnected):
                response = requests.get(
                    url,
                    auth=auth,
                    stream=True,
                    verify=verify,
                    timeout=limits.timeout,
                )
                if response.ok:
                    return response
        return response

    @staticmethod
    def download_raw(url: HttpUrl, auth: HTTPBasicAuth | None = None, verify: str | bool = True) -> bytes | None:
        """Download the raw content with the provided parameters.

        Arguments:
            url: The URL to access
            auth: The credentials needed to access the URL (default None)
            verify: The certificate needed to access the URL (default True)

        Returns:
            bytes | None: The raw content if the request is successful
        """
        response = HTTP.get(url, auth, verify)

        return response.content if response.ok else None

    @staticmethod
    def download(
        url: HttpUrl,
        auth: HTTPBasicAuth | None = None,
        verify: str | bool = True,
        encoding: str = "utf-8",
    ) -> str | None:
        """Download the content and decode with the provided parameters.

        Arguments:
            url: The URL to access
            auth: The credentials needed to access the URL (default None)
            verify: The certificate needed to access the URL (default True)
            encoding: The encoding to use when decoding the content (default "utf-8")

        Returns:
            str | None: The decoded content if the request is successful
        """
        with contextlib.suppress(ProxyError):
            if raw := HTTP.download_raw(url, auth, verify):
                return raw.decode(encoding)

        return None

    @staticmethod
    def download_compressed(
        url: HttpUrl,
        auth: HTTPBasicAuth | None = None,
        verify: str | bool = True,
        encoding: str = "utf-8",
    ) -> str | None:
        """Download the content, then decoompress and decode it with the provided parameters.

        Arguments:
            url: The URL to access
            auth: The credentials needed to access the URL (default None)
            verify: The certificate needed to access the URL (default True)
            encoding: The encoding to use when decoding the content (default "utf-8")

        Returns:
            str | None: The decompressed and decoded content if the request is successful
        """
        decompress: dict[str, Callable[[bytes], bytes]] = {
            "": lambda c: c,
            ".xz": lzma.decompress,
            ".gz": gzip.decompress,
            ".bzip2": bz2.decompress,
        }

        for suffix, method in decompress.items():
            with contextlib.suppress(ProxyError):
                if (raw := HTTP.download_raw(parse_obj_as(HttpUrl, f"{url}{suffix}"), auth, verify)) and (
                    output := method(raw).decode(encoding)
                ):
                    return output

        return None


def cache_file_outdated(file: Path) -> bool:
    """Check if the provided file is outdated.

    Arguments:
        file: File to check

    Returns:
        bool: True if the file is outdated, otherwise, false
    """
    tz = datetime.now().astimezone().tzinfo

    expiration_time = datetime.fromtimestamp(file.stat().st_mtime, tz) + ChopprShares.config.cache_timeout

    if outdated := expiration_time < datetime.now(tz):
        ChopprShares.log.debug(f"Outdated cache file will be refreshed: {file}")

    return outdated


def get_auth_and_verify(url: HttpUrl) -> tuple[HTTPBasicAuth | None, str | bool]:
    """Get the authentication credentials and certificate.

    Arguments:
        url: URL to get the credentials and certificate for

    Returns:
        tuple[HTTPBasicAuth | None, str | bool]: Credentials and certificate
    """
    if not url.host:
        return (None, True)

    auth = None
    credentials = Credentials.find(url.host)
    verify = ChopprShares.config.certificates.get(url.host, True)
    if credentials and isinstance(credentials.password, SecretStr):
        ChopprShares.log.debug(f"Found credentials for host: {url.host}")
        auth = HTTPBasicAuth(username=credentials.username, password=credentials.password.get_secret_value())
    else:
        ChopprShares.log.debug(f"No credentials for host: {url.host}")

    return (auth, verify)


def get_component_type(component: Component) -> PurlType | None:
    """Get the purl type for the provided component.

    Arguments:
        component: SBOM component

    Returns:
        PurlType | None: The purl type for the component
    """
    if component.purl and (match := re.match(r"pkg:(?P<type>.*?)/", component.purl)):
        with contextlib.suppress(KeyError):
            return PurlType(match["type"].upper())
    return None


def log_header(title: str) -> None:
    """Output a header to the log with the provided title, and padded with equal signs.

    Arguments:
        title: The header title
    """
    ChopprShares.log.info(f" {title} ".center(LOG_HEADER_LENGTH, "="))


def log_repo_pulls(expected: int, actual: int, purl_type: PurlType) -> None:
    """Log the success or failure of pulling a repository.

    Arguments:
        expected: The expected number of repositories
        actual: The number of repositories actually pulled
        purl_type: They repository type
    """
    match actual:
        case 0:
            ChopprShares.log.error(f"Failed to pull any {purl_type} repositories")
        case _ if actual == expected:
            ChopprShares.log.info(f"Successfully pulled all {purl_type} repositories")
        case _:
            ChopprShares.log.error(f"Failed to pull {expected - actual}/{expected} {purl_type} repositories")


def output_list(output_file: Path, items: list[str]) -> None:
    """Output list to the provided file.

    Arguments:
        output_file: Output file path
        items: List of strings to write to the file
    """
    with output_file.open("w") as file:
        file.writelines([f"{item}\n" for item in items])


def prepend_slash(text: str) -> str:
    """Prepend a slash to a string that doesn't have one.

    Arguments:
        text: A string to prepend a slash to

    Returns:
        str: The string with a prepended slash
    """
    return f"/{text.removeprefix('/')}"


def remove_parenthesis(text: str) -> str:
    """Remove text within parenthesis, to include empty parenthesis.

    Arguments:
        text: Text to remove parenthesis from

    Returns:
        str: Text with parenthesis removed
    """
    return re.sub(r"\(.*?\)", "", text)
